from datetime import date, datetime

from django.contrib.admin.models import ADDITION, CHANGE, DELETION, LogEntry
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.utils.timezone import localtime
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST

from import_export.admin import ImportMixin as BaseImportMixin
from import_export.results import RowResult

import json


class CustomImportMixin(BaseImportMixin):
    @method_decorator(require_POST)
    def process_import(self, request, **kwargs):
        """
        Perform the actual import action (after the user has confirmed the import)
        """
        if not self.has_import_permission(request):
            raise PermissionDenied

        confirm_form = self.create_confirm_form(request)
        if confirm_form.is_valid():
            import_formats = self.get_import_formats()
            input_format = import_formats[int(confirm_form.cleaned_data["format"])](
                encoding=self.from_encoding
            )
            encoding = None if input_format.is_binary() else self.from_encoding
            tmp_storage_cls = self.get_tmp_storage_class()
            tmp_storage = tmp_storage_cls(
                name=confirm_form.cleaned_data["import_file_name"],
                encoding=encoding,
                read_mode=input_format.get_read_mode(),
                **self.get_tmp_storage_class_kwargs(),
            )

            data = tmp_storage.read()
            dataset = input_format.create_dataset(data)

            # Store original values before import
            original_objects = {}
            res_kwargs = self.get_import_resource_kwargs(request=request, form=confirm_form, **kwargs)
            resource = self.choose_import_resource_class(confirm_form, request)(**res_kwargs)
            
            # Get a preview of the changes without committing
            dry_run_result = resource.import_data(
                dataset,
                dry_run=True,
                file_name=confirm_form.cleaned_data.get("original_file_name"),
                user=request.user,
            )
            
            # Save original values of objects that will be updated
            for row in dry_run_result.rows:
                if row.import_type == row.IMPORT_TYPE_UPDATE and row.object_id:
                    try:
                        obj = self.model.objects.get(pk=row.object_id)
                        # Store a copy of original values
                        original_values = {}
                        for field in self.model._meta.fields:
                            original_values[field.name] = getattr(obj, field.name)
                        original_objects[obj.pk] = original_values
                    except self.model.DoesNotExist:
                        pass
                    
            result = self.process_dataset(dataset, confirm_form, request, **kwargs)

            result.original_objects = original_objects

            tmp_storage.remove()

            return self.process_result(result, request)
        else:
            context = self.admin_site.each_context(request)
            context.update(
                {
                    "title": _("Import"),
                    "confirm_form": confirm_form,
                    "opts": self.model._meta,
                    "errors": confirm_form.errors,
                }
            )
            return TemplateResponse(request, [self.import_template_name], context)
        
    def _log_actions(self, result, request):
        """
        Create appropriate LogEntry instances for the result.
        """
        rows = {}
        for row in result:
            rows.setdefault(row.import_type, [])
            rows[row.import_type].append(row.instance)

        # Pass original_objects directly from result
        original_objects = getattr(result, 'original_objects', {})
        self._create_log_entries(request.user.pk, rows, original_objects)

    def _create_log_entries(self, user_pk, rows, original_objects):
        logentry_map = {
            RowResult.IMPORT_TYPE_NEW: ADDITION,
            RowResult.IMPORT_TYPE_UPDATE: CHANGE,
            RowResult.IMPORT_TYPE_DELETE: DELETION,
        }
        content_type = ContentType.objects.get_for_model(self.model)

        for import_type, instances in rows.items():
            if import_type in logentry_map:
                action_flag = logentry_map[import_type]
                for instance in instances:
                    # Reset change_message for each instance
                    change_message = ""
                    
                    # Use the stored original values for comparison
                    if import_type == RowResult.IMPORT_TYPE_UPDATE and hasattr(instance, 'pk') and instance.pk in original_objects:
                        original_values = original_objects[instance.pk]
                        
                        # Compare and find changes
                        changed_fields = []
                        for field in self.model._meta.fields:
                            field_name = field.name
                            if field_name in original_values:
                                original_value = original_values[field_name]
                                new_value = getattr(instance, field_name)

                                # Apply localtime for date/datetime fields
                                if isinstance(original_value, datetime):
                                    original_value = localtime(original_value).strftime('%Y-%m-%d %H:%M:%S')
                                if isinstance(new_value, datetime):
                                    new_value = localtime(new_value).strftime('%Y-%m-%d %H:%M:%S')
                                
                                # Format date objects for better display
                                if isinstance(original_value, date):
                                    original_value = original_value.strftime('%Y-%m-%d')
                                if isinstance(new_value, date):
                                    new_value = new_value.strftime('%Y-%m-%d')

                                if original_value != new_value:
                                    changed_fields.append(
                                        f"[{field.verbose_name}] \"{original_value}\" => \"{new_value}\""
                                    )
                        if changed_fields:
                            change_message = '[{"import_changed": {"fields": ' + json.dumps(changed_fields) + '}}]'
                    elif import_type == RowResult.IMPORT_TYPE_NEW:
                        change_message = '[{"added": {}}]'
                    elif import_type == RowResult.IMPORT_TYPE_DELETE:
                        change_message = '[{"deleted": {}}]'
                    else:
                        change_message = _("{} via custom import_export".format(import_type))

                    # Create log entry for EACH instance
                    LogEntry.objects.log_action(
                        user_id=user_pk,
                        content_type_id=content_type.pk,
                        object_id=instance.pk,
                        object_repr=str(instance),
                        action_flag=action_flag,
                        change_message=change_message,
                    )


from typing import Any

from django.contrib import admin
from django.contrib.admin.options import ModelAdmin
from django.contrib.admin.views.main import ChangeList
from django.core.validators import EMPTY_VALUES
from django.db.models import Model, QuerySet
from django.db.models.fields import (
    DateField,
    DateTimeField,
    Field,
)
from django.forms import ValidationError
from django.http import HttpRequest

from unfold.contrib.filters.forms import (
    RangeDateForm,
    RangeDateTimeForm,
)
from unfold.utils import parse_date_str, parse_datetime_str


class RangeDateFilter(admin.FieldListFilter):
    request = None
    parameter_name = None
    form_class = RangeDateForm
    template = "unfold/filters/filters_date_range.html"

    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)
        if not isinstance(field, DateField):
            raise TypeError(
                f"Class {type(self.field)} is not supported for {self.__class__.__name__}."
            )

        self.request = request
        if self.parameter_name is None:
            self.parameter_name = self.field_path

        if self.parameter_name + "_from" in params:
            value = params.pop(self.field_path + "_from")
            value = value[0] if isinstance(value, list) else value

            if value not in EMPTY_VALUES:
                self.used_parameters[self.field_path + "_from"] = value

        if self.parameter_name + "_to" in params:
            value = params.pop(self.field_path + "_to")
            value = value[0] if isinstance(value, list) else value

            if value not in EMPTY_VALUES:
                self.used_parameters[self.field_path + "_to"] = value

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        filters = {}

        value_from = self.used_parameters.get(self.parameter_name + "_from")
        if value_from not in EMPTY_VALUES:
            filters.update({self.parameter_name + "__date__gte": parse_date_str(value_from)})

        value_to = self.used_parameters.get(self.parameter_name + "_to")
        if value_to not in EMPTY_VALUES:
            filters.update({self.parameter_name + "__date__lte": parse_date_str(value_to)})

        try:
            return queryset.filter(**filters)
        except (ValueError, ValidationError):
            return None

    def expected_parameters(self) -> list[str]:
        return [
            f"{self.parameter_name}_from",
            f"{self.parameter_name}_to",
        ]

    def choices(self, changelist: ChangeList) -> tuple[dict[str, Any], ...]:
        return (
            {
                "request": self.request,
                "parameter_name": self.parameter_name,
                "form": self.form_class(
                    name=self.parameter_name,
                    data={
                        self.parameter_name + "_from": self.used_parameters.get(
                            self.parameter_name + "_from", None
                        ),
                        self.parameter_name + "_to": self.used_parameters.get(
                            self.parameter_name + "_to", None
                        ),
                    },
                ),
            },
        )


class RangeDateTimeFilter(admin.FieldListFilter):
    request = None
    parameter_name = None
    template = "unfold/filters/filters_datetime_range.html"
    form_class = RangeDateTimeForm

    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)
        if not isinstance(field, DateTimeField):
            raise TypeError(
                f"Class {type(self.field)} is not supported for {self.__class__.__name__}."
            )

        self.request = request
        if self.parameter_name is None:
            self.parameter_name = self.field_path

        if self.parameter_name + "_from_0" in params:
            value = params.pop(self.field_path + "_from_0")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_from_0"] = value

        if self.parameter_name + "_from_1" in params:
            value = params.pop(self.field_path + "_from_1")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_from_1"] = value

        if self.parameter_name + "_to_0" in params:
            value = params.pop(self.field_path + "_to_0")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_to_0"] = value

        if self.parameter_name + "_to_1" in params:
            value = params.pop(self.field_path + "_to_1")
            value = value[0] if isinstance(value, list) else value
            self.used_parameters[self.field_path + "_to_1"] = value

    def expected_parameters(self) -> list[str]:
        return [
            f"{self.parameter_name}_from_0",
            f"{self.parameter_name}_from_1",
            f"{self.parameter_name}_to_0",
            f"{self.parameter_name}_to_1",
        ]

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        filters = {}

        date_value_from = self.used_parameters.get(self.parameter_name + "_from_0")
        time_value_from = self.used_parameters.get(self.parameter_name + "_from_1")

        date_value_to = self.used_parameters.get(self.parameter_name + "_to_0")
        time_value_to = self.used_parameters.get(self.parameter_name + "_to_1")

        if date_value_from not in EMPTY_VALUES and time_value_from not in EMPTY_VALUES:
            filters.update(
                {
                    f"{self.parameter_name}__gte": parse_datetime_str(
                        f"{date_value_from} {time_value_from}"
                    ),
                }
            )

        if date_value_to not in EMPTY_VALUES and time_value_to not in EMPTY_VALUES:
            filters.update(
                {
                    f"{self.parameter_name}__lte": parse_datetime_str(
                        f"{date_value_to} {time_value_to}"
                    ),
                }
            )

        try:
            return queryset.filter(**filters)
        except (ValueError, ValidationError):
            return None

    def choices(self, changelist: ChangeList) -> tuple[dict[str, Any], ...]:
        return (
            {
                "request": self.request,
                "parameter_name": self.parameter_name,
                "form": self.form_class(
                    name=self.parameter_name,
                    data={
                        self.parameter_name + "_from_0": self.used_parameters.get(
                            self.parameter_name + "_from_0"
                        ),
                        self.parameter_name + "_from_1": self.used_parameters.get(
                            self.parameter_name + "_from_1"
                        ),
                        self.parameter_name + "_to_0": self.used_parameters.get(
                            self.parameter_name + "_to_0"
                        ),
                        self.parameter_name + "_to_1": self.used_parameters.get(
                            self.parameter_name + "_to_1"
                        ),
                    },
                ),
            },
        )



from dataclasses import dataclass
from typing import Callable, Optional, Union

from unfold.enums import ActionVariant

from .typing import ActionFunction


@dataclass(frozen=True)
class UnfoldAction:
    action_name: str
    method: ActionFunction
    description: str
    path: str
    attrs: Optional[dict] = None
    object_id: Optional[Union[int, str]] = None
    icon: Optional[str] = None
    variant: Optional[ActionVariant] = ActionVariant.DEFAULT


@dataclass
class Favicon:
    href: Union[str, Callable]
    rel: Optional[str] = None
    type: Optional[str] = None
    sizes: Optional[str] = None


@dataclass
class DropdownItem:
    title: str
    link: Union[str, Callable]
    icon: Optional[str] = None



import json
from django.contrib.admin.models import LogEntry, CHANGE
from django.contrib.contenttypes.models import ContentType
from django.utils.encoding import force_str

from adminsortable2.admin import SortableAdminMixin as BaseSortableAdminMixin

class SortableAdminMixin(BaseSortableAdminMixin):
    """
    Custom SortableAdminMixin that extends the functionality of adminsortable2's SortableAdminMixin.
    This mixin can be used to add sorting capabilities to Django admin models.
    """
    
    def update_order(self, request):
        """
        Override to log sorting actions in the admin log.
        """
        try:
            # 요청 데이터 추출
            data = json.loads(request.body)
            updated_items = data.get('updatedItems', [])
            
            # 변경 전 순서 정보 미리 저장
            old_orders = {}
            if updated_items:
                object_ids = [item[0] for item in updated_items]
                objects_before = self.model.objects.filter(id__in=object_ids)
                for obj in objects_before:
                    old_orders[str(obj.id)] = obj.order
            
            # 기본 정렬 처리 (부모 메서드 호출)
            response = super().update_order(request)
            
            # Admin Log에 변경 기록
            if updated_items and response.status_code == 200:
                content_type = ContentType.objects.get_for_model(self.model)
                
                # 변경된 항목의 IDs
                objects = self.model.objects.filter(id__in=[item[0] for item in updated_items])
                
                # 각 변경된 객체에 대한 메시지 생성
                for obj_id, new_order in updated_items:
                    obj = next((o for o in objects if str(o.id) == obj_id), None)
                    if obj:
                        # 원래 순서 확인
                        old_order = old_orders.get(obj_id, "알 수 없음")
                        
                        # 로그 메시지 생성 (원래 순서와 새 순서 모두 포함)
                        log_message = json.dumps([{
                            "changed": {
                                "fields": [f"order '{str(old_order)}' → '{str(new_order)}'"]
                            },
                        }])
                        
                        # Admin Log에 기록
                        LogEntry.objects.log_action(
                            user_id=request.user.id,
                            content_type_id=content_type.id,
                            object_id=obj_id,
                            object_repr=force_str(obj),
                            action_flag=CHANGE,
                            change_message=log_message
                        )
            
            return response
            
        except Exception as e:
            print(f"순서 변경 로깅 중 오류 발생: {e}")
            # 오류가 발생해도 기본 기능은 유지
            return super().update_order(request)